"""
Tools for working with still images and videos based on the ffmpeg library.
"""
import os
import subprocess as sp


def make_movie(stills_directory, input_file_format='*.png',
               movie_filename='', frames_per_second=30):
    """
    Make a movie out of a sequence of still frames.

    Parameters
    ----------
    frames_per_second : int
        The number of stills that get packed into one second of video.
    input_file_format : str
        The format of the still images that are read in.
    movie_filename : str
        The filename to apply to the video file.
        Default is empty string, in which case a filename is autogenerated..
    stills_directory : str
        The relative path to the directory that contains the still images.
    """
    # Generate a movie filename if one was not provided.
    if not movie_filename:
        movie_filename = ''.join((stills_directory, '.mp4'))

    # Prepare the arguments for the call to ffmpeg.
    input_file_pattern = ''.join(
        ['\'', os.path.join(stills_directory, input_file_format), '\''])
    codec = 'libx264'
    command = ' '.join(['ffmpeg -framerate', str(frames_per_second),
                        '-pattern_type glob -i', input_file_pattern,
                        '-c:v', codec, movie_filename])
    print command
    # shell=True is considered a bit of a security risk, but without it
    # this cammand won't parse properly
    sp.call(command, shell=True)


def break_movie(movie_filename, stills_directory, output_file_format='.jpg'):
    """
    Create a set of numbered still images from a movie.
    Run from the directory containing the movie file.

    Parameters
    ----------
    movie_directory : str
        The relative path to the directory that contains the video file.
    movie_filename : str
        The name of the video file.
    output_file_format : str
        The format of the output still images, indicated by the suffix.
    stills_directory
        The relative path to the directory that contains the still images.
    """
    # Prepare the arguments for the call to ffmpeg
    output_files = ''.join(['%05d', output_file_format])
    output_file_pattern = os.path.join(stills_directory, output_files)

    command = ' '.join(['ffmpeg -i', movie_filename, output_file_pattern])
    print command
    # shell=True is considered a bit of a security risk, but without it
    # this cammand won't parse properly
    sp.call(command, shell=True)



"""
One-dimensional grid task with noise

In this task, the agent has the challenge of discriminating between
actual informative state sensors, and a comparatively large number
of sensors that are pure noise distractors. Many learning methods
make the implicit assumption that all sensors are informative.
This task is intended to break them.
"""
from __future__ import print_function
import numpy as np

from beccatest.base_world import World as BaseWorld


class World(BaseWorld):
    """ 
    One-dimensional grid world with noise
    
    In this world, the agent steps forward and backward 
    along three positions on a line. The second position is rewarded 
    and the first and third positions are punished. Also, any actions 
    are penalized somewhat. It also includes some inputs that are pure noise. 
    Optimal performance is a reward of about .70 per time step.
    
    Attributes
    ----------
    action : array of floats
        The most recent set of action commands received. 
    brain_visualize_period : int
        The number of time steps between creating a full visualization of
        the ``brain``.
    energy_cost : float
        The punishment per position step taken.
    jump_fraction : float
        The fraction of time steps on which the agent jumps to 
        a random position.
    name : str
        A name associated with this world.
    name_long : str
        A longer name associated with this world.
    num_actions : int
        The number of action commands this world expects. This should be 
        the length of the action array received at each time step.
    num_noise_sensors : int
        Of the sensors, these are purely noise.
    num_real_sensors : int
        Of the sensors, these are the ones that represent position.
    num_sensors : int
        The number of sensor values the world returns to the brain 
        at each time step.
    reward_magnitude : float
        The magnitude of the reward and punishment given at 
        rewarded or punished positions.
    simple_state : int
        The nearest integer position of the agent in the world.
    world_state : float
        The actual position of the agent in the world. This can be fractional.
    world_visualize_period : int
        The number of time steps between creating visualizations of 
        the world.
    """
    def __init__(self, lifespan=None, test=False):
        """
        Set up the world 
        """
        BaseWorld.__init__(self, lifespan)
        self.reward_magnitude = 1.
        self.energy_cost = 0.01 * self.reward_magnitude
        self.jump_fraction = 0.1
        self.name = 'grid_1D_noise'
        self.name_long = 'noisy one dimensional grid world'
        print("Entering", self.name_long)
        self.num_real_sensors = 3
        # Number of sensors that have no basis in the world. 
        # These are noise meant to distract.
        self.num_noise_sensors = 0        
        self.num_sensors = self.num_noise_sensors + self.num_real_sensors
        self.num_actions = 2
        self.action = np.zeros(self.num_actions)
        self.world_state = 0      
        self.simple_state = 0       
        self.world_visualize_period = 1e3
        self.brain_visualize_period = 1e3


    def step(self, action): 
        """ 
        Take one time step through the world 

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        reward : float
            The amount of reward or punishment given by the world.
        sensors : array of floats
            The values of each of the sensors.
        """
        self.action = action.copy().ravel()
        self.action[np.nonzero(self.action)] = 1.
        self.timestep += 1 
        step_size = self.action[0] - self.action[1]

        # An approximation of metabolic energy
        energy = self.action[0] + self.action[1]
        self.world_state = self.world_state + step_size

        # At random intervals, jump to a random position in the world
        if np.random.random_sample() < self.jump_fraction:
            self.world_state = (self.num_real_sensors * 
                                np.random.random_sample())

        # Ensure that the world state falls between 0 and num_real_sensors 
        self.world_state -= (self.num_real_sensors * 
                             np.floor_divide(self.world_state, 
                                             self.num_real_sensors))
        self.simple_state = int(np.floor(self.world_state))

        # Assign sensors as zeros or ones. 
        # Represent the presence or absence of the current position in the bin.
        real_sensors = np.zeros(self.num_real_sensors)
        real_sensors[self.simple_state] = 1

        # Generate a set of noise sensors
        noise_sensors = np.round(np.random.random_sample(
            self.num_noise_sensors))
        sensors = np.hstack((real_sensors, noise_sensors))
        reward = -self.reward_magnitude
        if self.simple_state == 1:
            reward = self.reward_magnitude
        reward -= energy * self.energy_cost        

        return sensors, reward


    def visualize_world(self, brain):
        """ 
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_real_sensors + 
                               self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_real_sensors:self.num_real_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            state_image[self.num_real_sensors + 2 + action_index[0]] = 'x'
        print(''.join(state_image))



"""
One-dimensional grid delay task.

This task tests an agent's ability to properly ascribe reward to the
correct cause. The reward is delayed by a variable amount, which
makes the task challenging.
"""
from __future__ import print_function
import numpy as np

from beccatest.grid_1D import World as Grid_1D_World


class World(Grid_1D_World):
    """
    One-dimensional grid task with delayed reward

    This task is identical to the grid_1D task with the
    exception that reward is randomly delayed a few time steps.

    Attributes
    ----------
    future_reward : list of floats
        The reward that has been received, but will not be delivered to
        the agent yet. The index of the list indicates how many time
        steps will pass before delivery occurs.
    max_delay : int
        The maximum number of time steps that the reward may be delayed.
    name : str
        A name associated with this world.
    name_long : str
        A longer name associated with this world.
    world_visualize_period : int
        The number of time steps between creating visualizations of
        the world.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world. Base it on the grid_1D world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        Grid_1D_World.__init__(self, lifespan)
        self.name = 'grid_1D_delay'
        self.name_long = 'one dimensional grid world with delay'
        print('--delayed')
        self.max_delay = 1
        self.future_reward = [0.] * self.max_delay
        self.world_visualize_period = 1e4


    def assign_reward(self, sensors):
        """
        Calcuate the reward corresponding to the current state and assign
        it to a future time step.

        Parameters
        ----------
        sensors : array of floats
            The current sensor values.

        Returns
        -------
        reward : float
            The reward associated the set of input sensors.
        """
        new_reward = -sensors[8]
        new_reward += sensors[3]
        # Punish actions just a little
        new_reward -= self.energy  * self.energy_cost
        # Find the delay for the reward
        delay = np.random.randint(0, self.max_delay)
        #self.future_reward[delay] += 1. / float(self.max_delay)
        #self.future_reward[delay] = np.maximum(self.future_reward[delay], 1.)
        self.future_reward[delay] += new_reward
        # Advance the reward future by one time step
        self.future_reward.append(0.)
        reward = self.future_reward.pop(0)
        #print('    delay', delay, 'reward', reward)

        return reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_sensors + self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_sensors:self.num_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            for i in range(action_index.size):
                state_image[self.num_sensors + 2 + action_index[i]] = 'x'
        print(''.join(state_image))



"""
One-dimensional grid task.

This task tests an brain's ability to choose an appropriate action.
It is straightforward. Reward and punishment is clear and immediate.
There is only one reward state and it can be reached in a single
step.
"""
from __future__ import print_function
import numpy as np

from beccatest.base_world import World as BaseWorld

class World(BaseWorld):
    """
    One-dimensional grid world.

    In this task, the brain steps forward and backward along
    a nine-position line. The fourth position is rewarded and
    the ninth position is punished. There is also a slight
    punishment for effort expended in taking actions.
    Occasionally the brain will get
    involuntarily bumped to a random position on the line.
    This is intended to be a simple-as-possible
    task for troubleshooting BECCA.
    Optimal performance is a reward of about 90 per time step.

    Attributes
    ----------
    action : array of floats
        The most recent set of action commands received.
    brain_visualize_period : int
        The number of time steps between creating a full visualization of
        the ``brain``.
    energy_cost : float
        The punishment per position step taken.
    jump_fraction : float
        The fraction of time steps on which the agent jumps to
        a random position.
    name : str
        A name associated with this world.
    name_long : str
        A longer name associated with this world.
    num_actions : int
        The number of action commands this world expects. This should be
        the length of the action array received at each time step.
    num_sensors : int
        The number of sensor values the world returns to the brain
        at each time step.
    reward_magnitude : float
        The magnitude of the reward and punishment given at
        rewarded or punished positions.
    simple_state : int
        The nearest integer position of the agent in the world.
    world_state : float
        The actual position of the agent in the world. This can be fractional.
    world_visualize_period : int
        The number of time steps between creating visualizations of
        the world.
    """
    def __init__(self, lifespan=None):
        """
        Initialize the world.

        Parameters
        ----------
        lifespan : int
            The number of time steps to continue the world.
        """
        BaseWorld.__init__(self, lifespan)
        self.reward_magnitude = 1.
        self.energy_cost = self.reward_magnitude / 100.
        self.jump_fraction = 0.1
        self.name = 'grid_1D'
        self.name_long = 'one dimensional grid world'
        print("Entering", self.name_long)
        self.num_sensors = 9
        self.num_actions = 8
        self.action = np.zeros(self.num_actions)
        self.energy = 0.
        self.world_state = 0
        self.simple_state = 0
        self.world_visualize_period = 1e3
        self.brain_visualize_period = 1e3


    def step(self, action):
        """
        Advance the world one time step.

        Parameters
        ----------
        action : array of floats
            The set of action commands to execute.

        Returns
        -------
        self.reward : float
            The amount of reward or punishment given by the world.
        self.sensors : array of floats
            The values of each of the sensors.
        """
        self.action = action
        self.action = np.round(self.action)
        self.timestep += 1

        # Find the step size as combinations of the action commands
        #     action[i]     result
        #            0      1 step to the right
        #            1      2 steps to the right
        #            2      3 steps to the right
        #            3      4 steps to the right
        #            4      1 step to the left
        #            5      2 steps to the left
        #            6      3 steps to the left
        #            7      4 steps to the left
        step_size = (self.action[0] +
                     self.action[1] * 2 +
                     self.action[2] * 3 +
                     self.action[3] * 4 -
                     self.action[4] -
                     self.action[5] * 2 -
                     self.action[6] * 3 -
                     self.action[7] * 4)
        # Action cost is an approximation of metabolic energy.
        # Action cost is proportional to the number of steps taken.
        self.energy = (self.action[0] +
                       self.action[1] * 2 +
                       self.action[2] * 3 +
                       self.action[3] * 4 +
                       self.action[4] +
                       self.action[5] * 2 +
                       self.action[6] * 3 +
                       self.action[7] * 4)

        self.world_state = self.world_state + step_size

        # At random intervals, jump to a random position in the world.
        if np.random.random_sample() < self.jump_fraction:
            self.world_state = self.num_sensors * np.random.random_sample()

        # Ensure that the world state falls between 0 and 9.
        self.world_state -= self.num_sensors * np.floor_divide(
            self.world_state, self.num_sensors)
        self.simple_state = int(np.floor(self.world_state))
        if self.simple_state == 9:
            self.simple_state = 0

        # Represent the presence or absence of the current position in the bin.
        sensors = np.zeros(self.num_sensors)
        sensors[self.simple_state] = 1
        reward = self.assign_reward(sensors)

        return sensors, reward


    def assign_reward(self, sensors):
        """
        Calculate the total reward corresponding to the current state

        Parameters
        ----------
        sensors : array of floats
            The current sensor values.

        Returns
        -------
        reward : float
            The reward associated the set of input sensors.
        """
        reward = 0.
        reward -= sensors[8] * self.reward_magnitude
        reward += sensors[3] * self.reward_magnitude
        # Punish actions just a little
        reward -= self.energy  * self.energy_cost
        reward = np.maximum(reward, -self.reward_magnitude)

        return reward


    def visualize_world(self, brain):
        """
        Show what's going on in the world.
        """
        state_image = ['.'] * (self.num_sensors + self.num_actions + 2)
        state_image[self.simple_state] = 'O'
        state_image[self.num_sensors:self.num_sensors + 2] = '||'
        action_index = np.where(self.action > 0.1)[0]
        if action_index.size > 0:
            for i in range(action_index.size):
                state_image[self.num_sensors + 2 + action_index[i]] = 'x'
        print(''.join(state_image))

